@@ -1,7 +1,10 @@
 from __future__ import unicode_literals
-import urllib
 
 import csv
+import urllib
+import base64
+import hashlib
+import os.path
 
 from django.db import models
 from django import forms
@@ -15,6 +18,8 @@ from django.template.loader import render_to_string
 from django.utils.functional import cached_property
 from django.utils.encoding import python_2_unicode_compatible
 from django.template.defaultfilters import truncatechars
+from django.core.files import File
+from django.core.files.storage import default_storage
 
 from django_extensions.db.fields import UUIDField
 import html2text
@@ -70,6 +75,8 @@ class FormReponseHandler(FormSuccessHandler):
 @widgy.register
 class SaveDataHandler(FormSuccessHandler):
     editable = False
+    tooltip = _("Saves the data when the Form is filled out. This is enabled by"
+                " default.")
 
     class Meta:
         verbose_name = _('save data handler')
@@ -208,16 +215,6 @@ class WebToLeadMapperHandler(RepostHandler):
         return issubclass(cls, FieldMappingValue)
 
 
-def send_html_mail(subject, content, from_email, to):
-    msg = EmailMultiAlternatives(
-        subject=subject,
-        body=html2text.html2text(content),
-        from_email=from_email,
-        to=to)
-    msg.attach_alternative(content, 'text/html')
-    msg.send()
-
-
 class EmailSuccessHandlerBaseForm(forms.ModelForm):
     content = CKEditorField()
 
@@ -254,8 +251,21 @@ class EmailSuccessHandlerBase(StrDisplayNameMixin, FormSuccessHandler):
         })
 
     def execute(self, request, form):
-        message = self.format_message(request, form)
-        send_html_mail(self.subject, message, settings.SERVER_EMAIL, self.get_to_emails(form))
+        message_text = self.format_message(request, form)
+        msg = EmailMultiAlternatives(
+            subject=self.subject,
+            body=html2text.html2text(message_text),
+            from_email=settings.SERVER_EMAIL,
+            to=self.get_to_emails(form),
+        )
+        msg.attach_alternative(message_text, 'text/html')
+
+        if self.include_form_data:
+            for value in form.cleaned_data.values():
+                if isinstance(value, File):
+                    msg.attach(value.name, value.read(), getattr(value, 'content_type', None))
+
+        msg.send()
 
     def get_to_emails(self, form):
         raise NotImplemented
@@ -265,12 +275,19 @@ class EmailSuccessHandlerBase(StrDisplayNameMixin, FormSuccessHandler):
 class EmailSuccessHandler(EmailSuccessHandlerBase):
     to = models.EmailField(verbose_name=_('to'))
 
+    tooltip = _("This widget can be used to send yourself an email when a form"
+                " has been filled out. You can customize the body of the"
+                " email.")
+
     class Meta:
         verbose_name = _('admin success email')
         verbose_name_plural = _('admin success emails')
 
     def get_to_emails(self, form):
-        return [self.to]
+        if self.to:
+            return [self.to]
+        else:
+            return []
 
 # This is the only way to change a field provided by a base class.
 EmailSuccessHandler._meta.get_field('include_form_data').default = True
@@ -291,6 +308,8 @@ class EmailUserHandlerForm(EmailSuccessHandlerBaseForm):
 class EmailUserHandler(EmailSuccessHandlerBase):
     editable = True
     form = EmailUserHandlerForm
+    tooltip = _("This widget can be used to send the user an email when they"
+                " fill out a form. You can customize the body of the email.")
 
     # an input in our form
     to_ident = models.CharField(_('to'), max_length=36)
@@ -325,6 +344,8 @@ class EmailUserHandler(EmailSuccessHandlerBase):
 class SubmitButton(StrDisplayNameMixin, FormElement):
     text = models.CharField(max_length=255, default=lambda: ugettext('submit'), verbose_name=_('text'))
 
+    tooltip = _("The submit button for a form.")
+
     @property
     def deletable(self):
         return len([i for i in self.parent_form.depth_first_order() if isinstance(i, SubmitButton)]) > 1
@@ -415,6 +436,16 @@ class FormMeta(StrictDefaultChildrenMixin, Bucket):
         return isinstance(parent, Form)
 
 
+def friendly_uuid(uuid):
+    """
+    A shortend version of a UUID to use when collisions are acceptable.
+    The returned string will have 40 bits of entropy assuming a UUID4.
+    """
+    result = base64.b32encode(hashlib.sha1(str(uuid)).digest()[:5]).lower()
+    # avoid accidental profanity
+    return result.replace('e', '0').replace('o', '1').replace('u', '8').replace('i', '9')
+
+
 @widgy.register
 @python_2_unicode_compatible
 class Form(TabbedContainer, StrDisplayNameMixin, StrictDefaultChildrenMixin, Content):
@@ -432,6 +463,7 @@ class Form(TabbedContainer, StrDisplayNameMixin, StrictDefaultChildrenMixin, Con
         ('fields', FormBody, (), {}),
         ('meta', FormMeta, (), {}),
     ]
+    tooltip = _("Use this widget to build a Form that your users can fill out.")
 
     class FormQuerySet(QuerySet):
         def annotate_submission_count(self):
@@ -475,6 +507,16 @@ class Form(TabbedContainer, StrDisplayNameMixin, StrictDefaultChildrenMixin, Con
     def context_var(self):
         return 'form_instance_{node_pk}'.format(node_pk=self.node.pk)
 
+    @property
+    def success_key(self):
+        """
+        Used in the success URL to decide which form's success message
+        to show. Collisions are OK because there aren't expected to be
+        many forms on one page, and a collision will only result in both
+        forms' success messages being shown.
+        """
+        return friendly_uuid(self.ident)
+
     def action_url(self, widgy, tried=()):
         # the `tried` argument is just for nice error reporting
         assert widgy, 'get_form_action_url not found in any owners. Tried: %s' % (tried,)
@@ -501,11 +543,11 @@ class Form(TabbedContainer, StrDisplayNameMixin, StrictDefaultChildrenMixin, Con
         ctx = {
             'form': form,
             'action_url': action_url,
-            'success': request and request.GET.get('success') == str(self.node.pk),
+            'success': request and request.GET.get('success') == self.success_key,
         }
 
         with update_context(context, ctx):
-            return super(Form, self).render(context, template='widgy/form_builder/form/render.html')
+            return super(Form, self).render(context)
 
     def execute(self, request, form):
         # TODO: does this redirect belong here or in the view?  We populate
@@ -513,7 +555,7 @@ class Form(TabbedContainer, StrDisplayNameMixin, StrictDefaultChildrenMixin, Con
         # the model too.
         resp = redirect(build_url(
             request.GET['from'],
-            success=self.node.pk,
+            success=self.success_key,
         ))
         for child in self.depth_first_order():
             if isinstance(child, FormReponseHandler):
@@ -538,7 +580,7 @@ class Form(TabbedContainer, StrDisplayNameMixin, StrictDefaultChildrenMixin, Con
         """
         A dictionary of formfield name -> FormField widget
         """
-        ret = {}
+        ret = SortedDict()
         for child in self.depth_first_order():
             if isinstance(child, FormField):
                 ret[child.get_formfield_name()] = child
@@ -641,6 +683,13 @@ class FormField(StrDisplayNameMixin, BaseFormField):
         """
         return unicode(value)
 
+    @property
+    def widget_attrs(self):
+        attrs = {}
+        if self.required:
+            attrs.update(required='required')
+        return attrs
+
 
 class FormInputForm(FormFieldForm):
     class Meta:
@@ -677,44 +726,48 @@ class FormInput(FormField):
 
     type = models.CharField(choices=FORM_INPUT_TYPES, max_length=255, verbose_name=_('type'))
 
+    tooltip = _("A single form input. This widget can accept different types of"
+                " input like text, number, or email. See a more complete list"
+                " inside the widget.")
+
     class Meta:
         verbose_name = _('form input')
         verbose_name_plural = _('form inputs')
 
+    @property
+    def widget_attrs(self):
+        attrs = super(FormInput, self).widget_attrs
+        attrs['type'] = self.type
+
+        if self.type == 'date':
+            # Use type text because Kalendae doesn't play well with type=date
+            attrs['type'] = 'text'
+            attrs['class'] = 'date auto-kal'
+        return attrs
+
     @property
     def formfield_class(self):
         return self.FORMFIELD_CLASSES.get(self.type, forms.CharField)
 
     @property
     def widget(self):
-        attrs = {
-            'type': self.type,
-        }
-        if self.required:
-            attrs['required'] = 'required'
-
-        if self.type == 'date':
-            # Use type text because Kalendae doesn't play well with type=date
-            attrs['type'] = 'text'
-            attrs['class'] = 'date auto-kal'
-
-        return forms.TextInput(attrs=attrs)
+        return self.formfield_class.widget(attrs=self.widget_attrs)
 
 
 @widgy.register
 class Textarea(FormField):
     formfield_class = forms.CharField
 
+    tooltip = _("Add this to your form to allow users to add large amounts of"
+                " text.")
+
     class Meta:
         verbose_name = _('text area')
         verbose_name_plural = _('text areas')
 
     @property
     def widget(self):
-        attrs = {}
-        if self.required:
-            attrs['required'] = 'required'
-        return forms.Textarea(attrs=attrs)
+        return forms.Textarea(attrs=self.widget_attrs)
 
 
 class BaseChoiceField(FormField):
@@ -734,15 +787,7 @@ class BaseChoiceField(FormField):
 
     @property
     def widget(self):
-        attrs = self.widget_attrs
-        return self.widget_class(attrs=attrs)
-
-    @property
-    def widget_attrs(self):
-        attrs = {}
-        if self.required:
-            attrs.update(required='required')
-        return attrs
+        return self.widget_class(attrs=self.widget_attrs)
 
 
 @widgy.register
@@ -761,6 +806,10 @@ class ChoiceField(BaseChoiceField):
         ('radios', _('Radio buttons')),
     ])
 
+    tooltip = _("Use this to allow your users to choose from a list of"
+                " options. ChoiceFields can be displayed as radio buttons"
+                " or as a dropdown.")
+
     def get_choices(self):
         choices = super(ChoiceField, self).get_choices()
         if self.type == 'select':
@@ -786,6 +835,10 @@ class MultipleChoiceField(BaseChoiceField):
         ('select', _('Multi-select')),
     ])
 
+    tooltip = _("Use this to allow your users to choose one or more options"
+                " from a list of choices. This field can be displayed as"
+                " checkboxes or as a multi-select dropdown.")
+
     @property
     def widget_attrs(self):
         attrs = super(MultipleChoiceField, self).widget_attrs
@@ -814,6 +867,9 @@ class Uncaptcha(BaseFormField):
     editable = False
     formfield_class = forms.CharField
 
+    tooltip = _("This widget uses a special script to detect and prevent"
+                " spam.")
+
     def get_form_mixins(self):
         # since validating an uncaptcha widget requires access to the
         # csrfmiddlewaretoken and not just our field value, create a
@@ -844,6 +900,34 @@ class Uncaptcha(BaseFormField):
         verbose_name_plural = _('uncaptchas')
 
 
+@widgy.register
+class FileUpload(FormField):
+    formfield_class = forms.FileField
+    storage = default_storage
+
+    def generate_filename(self, filename):
+        return os.path.join(
+            'form-uploads',
+            self.storage.get_valid_name(os.path.basename(filename))
+        )
+
+    def serialize_value(self, value):
+        if value:
+            filename = self.generate_filename(value.name)
+            filename = self.storage.save(filename, value)
+            return self.storage.url(filename)
+        else:
+            return ''
+
+
+@widgy.register
+class ImageUpload(FileUpload):
+    formfield_class = forms.ImageField
+
+    class Meta:
+        proxy = True
+
+
 class FormSubmission(models.Model):
     """
     Holds the data from one submission of a Form.
